import argparse
import glob
import os

import isort
import nbformat


def sort_imports(file):
    print("Sorting imports in notebook {}".format(file))

    # Read notebook file
    notebook: nbformat.NotebookNode = nbformat.read(file, nbformat.NO_CONVERT)

    for cell in notebook.cells:
        # Only format "code" cells
        if cell.cell_type != "code":
            continue
        # Get source code of cell and format using isort
        source = cell.source
        source_as_string = "".join(source)
        source_as_string_formatted = isort.code(source_as_string)
        # Convert formatted code to list and overwrite JSON data
        source_formatted = source_as_string_formatted.strip("\n").splitlines(True)
        cell.source = source_formatted

    # Write notebook file back
    nbformat.write(notebook, file)


def clear_output(file):
    print("Clearing output in notebook {}".format(file))

    # Read notebook file
    notebook: nbformat.NotebookNode = nbformat.read(file, nbformat.NO_CONVERT)

    for cell in notebook.cells:
        # Only format "code" cells
        if cell.cell_type != "code":
            continue
        # Clear execution count and output
        cell.execution_count = None
        cell.outputs = []

    # Write notebook file back
    nbformat.write(notebook, file)


def main():
    # Parse input arguments
    parser = argparse.ArgumentParser(description="Run `isort` on Jupyter notebooks.")
    parser.add_argument(
        "paths",
        nargs="+",
        metavar="INPUT_PATH",
        help="One or more notebook files or directories",
    )
    parser.add_argument(
        "-i", "--sort-imports", action="store_true", help="Sort imports"
    )
    parser.add_argument(
        "-c", "--clear-output", action="store_true", help="Clear output"
    )
    args = parser.parse_args()

    if not (args.sort_imports or args.clear_output):
        print("No action requested, specify at least '-i' or '-c'")
        return

    for path in args.paths:
        if os.path.isfile(path):
            files = [path]
        elif os.path.isdir(path):
            files_rel_path = glob.glob("**/*.ipynb", root_dir=path, recursive=True)
            files = [os.path.join(path, file) for file in files_rel_path]
        else:
            files = []

        for file in files:
            if args.sort_imports:
                sort_imports(file)
            if args.clear_output:
                clear_output(file)


if __name__ == "__main__":
    main()
